const ProtocolAction = require('./_base-action.js');

/**
 * Change focus to another window. The window to change focus to may be specified by its server assigned window handle, or by the value of its name attribute.
 *
 * To find out the window handle use `windowHandles` command
 *
 * @example
 *  this.demoTest = function (browser) {
 *    browser.switchToMainWindow(function(result) {
 *
 *    });
 *  };
 *
 *  this.demoTestAsync = async function (browser) {
 *    const result = await browser.switchToMainWindow();
 *  };
 *
 *
 * @method switchToMainWindow
 * @param {string} handleOrName The server assigned window handle or the name attribute.
 * @param {function} [callback] Optional callback function to be called when the command finishes.
 * @see window
 * @api protocol.contexts
 */
module.exports = class Action extends ProtocolAction {
  command(callback) {
    return this.transportActions.getAllWindowHandles((result) => {
      const handle = result.value[0];

      this.transportActions.switchToWindow(handle, callback);
    });
  }
};
